from __future__ import annotations

from collections.abc import Iterable, Mapping
from typing import Any

TEXT_VERSION = "v1"


def sanitize_model_id(model_id: str) -> str:
    return model_id.replace("/", "__").replace(":", "_")


def _get_field(obj: Any, names: Iterable[str]) -> Any:
    if isinstance(obj, Mapping):
        for name in names:
            if name in obj and obj[name] not in (None, ""):
                return obj[name]
        return None

    for name in names:
        value = getattr(obj, name, None)
        if value not in (None, ""):
            return value
    return None


def _as_text(value: Any) -> str:
    if value is None:
        return ""
    return str(value).strip()


def airport_to_text(airport: Any) -> str:
    parts = [
        _as_text(_get_field(airport, ["name"])),
        _as_text(_get_field(airport, ["municipality", "city"])),
        _as_text(_get_field(airport, ["iso_country", "country"])),
        _as_text(_get_field(airport, ["iso_region", "region"])),
        _as_text(_get_field(airport, ["iata_code", "iata"])),
        _as_text(_get_field(airport, ["gps_code", "icao_code", "icao"])),
        _as_text(_get_field(airport, ["ident"])),
        _as_text(_get_field(airport, ["local_code"])),
        _as_text(_get_field(airport, ["keywords"])),
    ]
    return " ".join(part for part in parts if part)


def airport_to_record(airport: Any) -> dict[str, Any]:
    return {
        "id": _get_field(airport, ["id"]),
        "ident": _get_field(airport, ["ident"]),
        "name": _get_field(airport, ["name"]),
        "iata": _get_field(airport, ["iata_code", "iata"]),
        "icao": _get_field(airport, ["gps_code", "icao_code", "icao"]),
        "country": _get_field(airport, ["iso_country", "country"]),
        "region": _get_field(airport, ["iso_region", "region"]),
        "municipality": _get_field(airport, ["municipality", "city"]),
        "latitude": _get_field(airport, ["latitude_deg", "lat", "latitude"]),
        "longitude": _get_field(airport, ["longitude_deg", "lon", "longitude"]),
        "type": _get_field(airport, ["type"]),
    }


def coerce_airports(airports: Any) -> list[Any]:
    if airports is None:
        return []

    if hasattr(airports, "iterrows"):
        return [row for _, row in airports.iterrows()]

    if isinstance(airports, Mapping):
        return [airports]

    if isinstance(airports, (str, bytes)):
        return [airports]

    if isinstance(airports, Iterable):
        return list(airports)

    return [airports]


def try_import_pandas():
    try:
        import pandas as pd
    except ImportError:
        return None
    return pd



import heapq
from collections.abc import Sequence
from typing import Optional

from ..core.distance import distance
from ..core.loader import get_airport_by_iata, get_airport_by_icao, get_all_airports
from ..exceptions import RoutingError
from ..models.airport import Airport
from ..utils.constants import DEFAULT_CRUISE_SPEED_KTS, DEFAULT_MAX_LEG_KM
from ..utils.logging import get_logger
from ..utils.units import DistanceUnit, convert_distance

logger = get_logger()


def estimate_flight_time_hours(
    from_airport: Airport,
    to_airport: Airport,
    speed_kts: float = DEFAULT_CRUISE_SPEED_KTS,
    model: str = "haversine",
) -> float:
    dist_nmi = distance(
        from_airport.latitude_deg,
        from_airport.longitude_deg,
        to_airport.latitude_deg,
        to_airport.longitude_deg,
        model=model,
        unit="nmi",
    )

    return dist_nmi / speed_kts


def estimate_flight_time_h_m(
    from_airport: Airport,
    to_airport: Airport,
    speed_kts: float = DEFAULT_CRUISE_SPEED_KTS,
    model: str = "haversine",
) -> tuple[int, int]:
    hours_total = estimate_flight_time_hours(from_airport, to_airport, speed_kts, model)
    hours = int(hours_total)
    minutes = int((hours_total - hours) * 60)

    return (hours, minutes)


def route_distance(
    airports: Sequence[Airport],
    model: str = "haversine",
    unit: DistanceUnit = "km",
) -> float:
    if len(airports) < 2:
        return 0.0

    total_dist_km = 0.0

    for i in range(len(airports) - 1):
        a1 = airports[i]
        a2 = airports[i + 1]

        dist_km = distance(
            a1.latitude_deg,
            a1.longitude_deg,
            a2.latitude_deg,
            a2.longitude_deg,
            model=model,
            unit="km",
        )

        total_dist_km += dist_km

    return convert_distance(total_dist_km, "km", unit)


def route_distance_by_codes(
    codes: Sequence[str],
    code_type: str = "iata",
    model: str = "haversine",
    unit: DistanceUnit = "km",
) -> float:
    airports = []

    for code in codes:
        if code_type == "iata":
            airport = get_airport_by_iata(code)
        elif code_type == "icao":
            airport = get_airport_by_icao(code)
        else:
            airport = get_airport_by_iata(code)
            if airport is None:
                airport = get_airport_by_icao(code)

        if airport is None:
            raise RoutingError(f"Airport not found: {code}")

        airports.append(airport)

    return route_distance(airports, model, unit)


def shortest_path(
    origin_code: str,
    destination_code: str,
    code_type: str = "iata",
    max_leg_km: float = DEFAULT_MAX_LEG_KM,
    allowed_types: Optional[Sequence[str]] = None,
    avoid_countries: Optional[Sequence[str]] = None,
) -> list[Airport]:
    if code_type == "iata":
        origin = get_airport_by_iata(origin_code)
        destination = get_airport_by_iata(destination_code)
    elif code_type == "icao":
        origin = get_airport_by_icao(origin_code)
        destination = get_airport_by_icao(destination_code)
    else:
        origin = get_airport_by_iata(origin_code) or get_airport_by_icao(origin_code)
        destination = get_airport_by_iata(destination_code) or get_airport_by_icao(destination_code)

    if origin is None:
        raise RoutingError(f"Origin airport not found: {origin_code}")
    if destination is None:
        raise RoutingError(f"Destination airport not found: {destination_code}")

    all_airports = get_all_airports()

    if allowed_types:
        all_airports = [a for a in all_airports if a.type in allowed_types]

    if avoid_countries:
        avoid_set = {c.upper() for c in avoid_countries}
        all_airports = [a for a in all_airports if a.iso_country not in avoid_set]

    def airport_key(airport: Airport) -> tuple:
        if airport.id is not None:
            return ("id", airport.id)
        if airport.ident:
            return ("ident", airport.ident)
        if airport.gps_code:
            return ("gps", airport.gps_code)
        if airport.iata_code:
            return ("iata", airport.iata_code)
        return ("coords", airport.latitude_deg, airport.longitude_deg, airport.name)

    airport_map = {airport_key(a): a for a in all_airports}
    origin_key = airport_key(origin)
    destination_key = airport_key(destination)

    if origin_key not in airport_map or destination_key not in airport_map:
        raise RoutingError("Origin or destination excluded by filters")

    def heuristic(a: Airport, b: Airport) -> float:
        return distance(
            a.latitude_deg,
            a.longitude_deg,
            b.latitude_deg,
            b.longitude_deg,
            model="haversine",
            unit="km",
        )

    pq = [(0.0, 0.0, origin_key, [origin])]
    visited = set()

    while pq:
        f_score, g_score, current_key, path = heapq.heappop(pq)

        if current_key in visited:
            continue

        visited.add(current_key)
        current = airport_map[current_key]

        if current_key == destination_key:
            return path

        if len(visited) > 1000:
            logger.warning("Shortest path search exceeded 1000 visited nodes, stopping")
            break

        for neighbor in all_airports:
            neighbor_key = airport_key(neighbor)
            if neighbor_key in visited:
                continue

            leg_dist = heuristic(current, neighbor)

            if leg_dist > max_leg_km:
                continue

            new_g_score = g_score + leg_dist
            new_f_score = new_g_score + heuristic(neighbor, destination)

            heapq.heappush(pq, (new_f_score, new_g_score, neighbor_key, path + [neighbor]))

    raise RoutingError(
        f"No path found from {origin_code} to {destination_code} with given constraints"
    )



import importlib
import sys
from pathlib import Path


def _reload_config(monkeypatch, env):
    keys = [
        "HF_TOKEN",
        "HF_API_TOKEN",
        "HUGGINGFACE_HUB_TOKEN",
        "AERONAVX_CACHE",
        "AERONAVX_EMBED_MODEL",
        "AERONAVX_OFFLINE",
    ]
    for key in keys:
        monkeypatch.delenv(key, raising=False)

    for key, value in env.items():
        monkeypatch.setenv(key, value)

    if "aeronavx.hf.config" in sys.modules:
        del sys.modules["aeronavx.hf.config"]

    import aeronavx.hf.config as config

    return importlib.reload(config)


def test_token_precedence(monkeypatch):
    config = _reload_config(
        monkeypatch,
        {
            "HF_API_TOKEN": "api-token",
            "HUGGINGFACE_HUB_TOKEN": "hub-token",
        },
    )
    assert config.HF_TOKEN == "api-token"

    config = _reload_config(
        monkeypatch,
        {
            "HF_TOKEN": "primary-token",
            "HF_API_TOKEN": "api-token",
            "HUGGINGFACE_HUB_TOKEN": "hub-token",
        },
    )
    assert config.HF_TOKEN == "primary-token"


def test_cache_dir_override(monkeypatch, tmp_path):
    config = _reload_config(monkeypatch, {"AERONAVX_CACHE": str(tmp_path)})
    assert config.CACHE_DIR == Path(tmp_path)


def test_embed_model_override(monkeypatch):
    config = _reload_config(monkeypatch, {"AERONAVX_EMBED_MODEL": "custom/model"})
    assert config.EMBED_MODEL == "custom/model"


def test_offline_flag(monkeypatch):
    config = _reload_config(monkeypatch, {"AERONAVX_OFFLINE": "1"})
    assert config.USE_LOCAL_ONLY is True

    config = _reload_config(monkeypatch, {"AERONAVX_OFFLINE": "0"})
    assert config.USE_LOCAL_ONLY is False



import logging
import os
import time
from collections import defaultdict

from fastapi import Depends, FastAPI, HTTPException, Query, Request
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import JSONResponse

from ..core.airports import get
from ..core.distance import distance
from ..core.emissions import estimate_co2_kg_by_codes
from ..core.emissions_advanced import (
    AircraftType,
    FuelType,
    calculate_flight_emissions,
    compare_saf_savings,
)
from ..core.network_intelligence import identify_global_hubs
from ..core.passenger_experience import calculate_jet_lag
from ..core.routing import estimate_flight_time_hours
from ..core.search import nearest_airports, search_airports_by_name
from ..core.synthetic_routes import generate_route
from ..exceptions import AeroNavXError

logger = logging.getLogger("aeronavx.api")

app = FastAPI(title="AeroNavX API", description="AI aviation SDK API", version="3.1.0")

# --- CORS ---
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_methods=["*"],
    allow_headers=["*"],
)

# --- Rate Limiting (in-memory, sliding window) ---
_rate_limit_store: dict[str, list[float]] = defaultdict(list)
_RATE_LIMIT = int(os.environ.get("AERONAVX_RATE_LIMIT", "60"))


def _check_rate_limit(client_ip: str) -> None:
    now = time.time()
    window_start = now - 60.0
    timestamps = _rate_limit_store[client_ip]
    _rate_limit_store[client_ip] = [t for t in timestamps if t > window_start]
    if len(_rate_limit_store[client_ip]) >= _RATE_LIMIT:
        raise HTTPException(status_code=429, detail="Too many requests. Try again later.")
    _rate_limit_store[client_ip].append(now)


# --- API Key Authentication ---
_API_KEY = os.environ.get("AERONAVX_API_KEY")


def _verify_api_key(request: Request) -> None:
    if _API_KEY is None:
        return
    key = request.headers.get("X-API-Key")
    if key != _API_KEY:
        raise HTTPException(status_code=401, detail="Invalid or missing API key.")


# --- Request logging + rate limit middleware ---
@app.middleware("http")
async def request_middleware(request: Request, call_next):
    start = time.time()
    path = request.url.path

    # Rate limiting (skip /health)
    if path != "/health":
        client_ip = request.client.host if request.client else "unknown"
        try:
            _check_rate_limit(client_ip)
        except HTTPException as exc:
            return JSONResponse(
                status_code=exc.status_code,
                content={"detail": exc.detail},
            )

    response = await call_next(request)
    duration_ms = (time.time() - start) * 1000
    logger.info(
        "%s %s → %s (%.1fms)",
        request.method,
        path,
        response.status_code,
        duration_ms,
    )
    return response


# --- Semantic search engine (lazy loaded) ---
_semantic_engine = None


def _get_semantic_engine():
    global _semantic_engine
    if _semantic_engine is None:
        try:
            from ..hf.semantic_search import SemanticAirportSearch
        except ImportError as exc:
            raise HTTPException(
                status_code=501,
                detail=(
                    "Semantic search requires the HF extra. "
                    "Install with `pip install aeronavx[hf]`."
                ),
            ) from exc
        from ..core.loader import get_all_airports

        _semantic_engine = SemanticAirportSearch(get_all_airports())
    return _semantic_engine


# --- Endpoints ---


@app.get("/health")
async def health():
    return {"status": "ok", "service": "AeroNavX API", "version": "3.1.0"}


@app.get("/airport/{code}")
async def get_airport(
    code: str,
    code_type: str = Query("auto", pattern="^(iata|icao|auto)$"),
    _auth: None = Depends(_verify_api_key),
):
    try:
        airport = get(code, code_type=code_type)

        if airport is None:
            raise HTTPException(status_code=404, detail=f"Airport not found: {code}")

        return airport.as_dict()

    except AeroNavXError as e:
        raise HTTPException(status_code=400, detail=str(e))


@app.get("/distance")
async def calculate_distance(
    from_code: str = Query(..., alias="from"),
    to_code: str = Query(..., alias="to"),
    code_type: str = Query("auto", pattern="^(iata|icao|auto)$"),
    model: str = Query("haversine", pattern="^(haversine|slc|vincenty)$"),
    unit: str = Query("km", pattern="^(km|mi|nmi)$"),
    _auth: None = Depends(_verify_api_key),
):
    try:
        from_airport = get(from_code, code_type=code_type)
        to_airport = get(to_code, code_type=code_type)

        if from_airport is None:
            raise HTTPException(status_code=404, detail=f"Origin airport not found: {from_code}")

        if to_airport is None:
            raise HTTPException(status_code=404, detail=f"Destination airport not found: {to_code}")

        dist = distance(
            from_airport.latitude_deg,
            from_airport.longitude_deg,
            to_airport.latitude_deg,
            to_airport.longitude_deg,
            model=model,
            unit=unit,
        )

        return {
            "from": from_airport.as_dict(),
            "to": to_airport.as_dict(),
            "distance": dist,
            "unit": unit,
            "model": model,
        }

    except AeroNavXError as e:
        raise HTTPException(status_code=400, detail=str(e))


@app.get("/nearest")
async def find_nearest(
    lat: float = Query(..., ge=-90, le=90),
    lon: float = Query(..., ge=-180, le=180),
    n: int = Query(5, ge=1, le=100),
    _auth: None = Depends(_verify_api_key),
):
    try:
        airports = nearest_airports(lat, lon, n=n)

        return {
            "query": {"lat": lat, "lon": lon},
            "count": len(airports),
            "airports": [a.as_dict() for a in airports],
        }

    except AeroNavXError as e:
        raise HTTPException(status_code=400, detail=str(e))


@app.get("/search")
async def search(
    q: str = Query(..., min_length=1),
    limit: int = Query(20, ge=1, le=100),
    _auth: None = Depends(_verify_api_key),
):
    try:
        airports = search_airports_by_name(q, limit=limit)

        return {"query": q, "count": len(airports), "airports": [a.as_dict() for a in airports]}

    except AeroNavXError as e:
        raise HTTPException(status_code=400, detail=str(e))


@app.get("/flight-time")
async def flight_time(
    from_code: str = Query(..., alias="from"),
    to_code: str = Query(..., alias="to"),
    speed_kts: float = Query(450.0, ge=100, le=1000),
    _auth: None = Depends(_verify_api_key),
):
    try:
        from_airport = get(from_code, code_type="auto")
        to_airport = get(to_code, code_type="auto")

        if from_airport is None:
            raise HTTPException(status_code=404, detail=f"Origin airport not found: {from_code}")

        if to_airport is None:
            raise HTTPException(status_code=404, detail=f"Destination airport not found: {to_code}")

        time_hours = estimate_flight_time_hours(from_airport, to_airport, speed_kts=speed_kts)

        hours = int(time_hours)
        minutes = int((time_hours - hours) * 60)

        return {
            "from": from_airport.as_dict(),
            "to": to_airport.as_dict(),
            "speed_kts": speed_kts,
            "time_hours": time_hours,
            "time_formatted": f"{hours}h {minutes}m",
        }

    except AeroNavXError as e:
        raise HTTPException(status_code=400, detail=str(e))


@app.get("/emissions")
async def emissions(
    from_code: str = Query(..., alias="from"),
    to_code: str = Query(..., alias="to"),
    model: str = Query("haversine", pattern="^(haversine|slc|vincenty)$"),
    _auth: None = Depends(_verify_api_key),
):
    try:
        co2_kg = estimate_co2_kg_by_codes(from_code, to_code, code_type="auto", model=model)

        from_airport = get(from_code, code_type="auto")
        to_airport = get(to_code, code_type="auto")

        return {
            "from": from_airport.as_dict(),
            "to": to_airport.as_dict(),
            "co2_kg_per_passenger": co2_kg,
            "model": model,
        }

    except AeroNavXError as e:
        raise HTTPException(status_code=400, detail=str(e))


@app.get("/semantic-search")
async def semantic_search(
    q: str = Query(..., min_length=1),
    top_k: int = Query(5, ge=1, le=50),
    _auth: None = Depends(_verify_api_key),
):
    try:
        engine = _get_semantic_engine()
        results = engine.search(q, top_k=top_k, return_format="list")
        return {"query": q, "count": len(results), "results": results}
    except HTTPException:
        raise
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))


@app.get("/jet-lag")
async def jet_lag(
    from_code: str = Query(..., alias="from"),
    to_code: str = Query(..., alias="to"),
    age: int = Query(30, ge=1, le=120),
    _auth: None = Depends(_verify_api_key),
):
    try:
        origin = get(from_code, code_type="auto")
        destination = get(to_code, code_type="auto")

        if origin is None or destination is None:
            raise HTTPException(status_code=404, detail="Airport not found")

        result = calculate_jet_lag(origin, destination, age=age)
        return {
            "from": origin.as_dict(),
            "to": destination.as_dict(),
            "timezone_difference_hours": result.timezone_difference_hours,
            "direction": result.direction.value,
            "severity": result.severity.value,
            "estimated_recovery_days": result.estimated_recovery_days,
        }
    except HTTPException:
        raise
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))


@app.get("/hubs")
async def hubs(
    top_n: int = Query(10, ge=1, le=50),
    include_unscheduled: bool = Query(False),
    _auth: None = Depends(_verify_api_key),
):
    try:
        results = identify_global_hubs(top_n=top_n, scheduled_service_only=not include_unscheduled)
        return {
            "count": len(results),
            "hubs": [
                {
                    "airport": hub.airport.as_dict(),
                    "hub_score": hub.hub_score,
                    "connectivity_score": hub.connectivity_score,
                    "type_score": hub.type_score,
                }
                for hub in results
            ],
        }
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))


@app.get("/emissions-advanced")
async def emissions_advanced(
    from_code: str = Query(..., alias="from"),
    to_code: str = Query(..., alias="to"),
    aircraft_type: str = Query("narrow_body"),
    fuel_type: str = Query("jet_a1"),
    load_factor: float = Query(0.85, ge=0.1, le=1.0),
    saf_percent: float = Query(0.0, ge=0.0, le=100.0),
    _auth: None = Depends(_verify_api_key),
):
    try:
        emissions = calculate_flight_emissions(
            from_code,
            to_code,
            aircraft_type=AircraftType(aircraft_type),
            fuel_type=FuelType(fuel_type),
            load_factor=load_factor,
            code_type="auto",
        )
        response = {
            "from": from_code,
            "to": to_code,
            "distance_km": emissions.distance_km,
            "fuel_kg": emissions.fuel_kg,
            "total_co2_kg": emissions.total_co2_kg,
            "co2_per_passenger_kg": emissions.co2_per_passenger_kg,
            "aircraft_type": emissions.aircraft_type.value,
            "fuel_type": emissions.fuel_type.value,
            "load_factor": emissions.load_factor,
        }
        if saf_percent > 0:
            savings = compare_saf_savings(
                from_code,
                to_code,
                saf_percentage=saf_percent,
                aircraft_type=AircraftType(aircraft_type),
                load_factor=load_factor,
                code_type="auto",
            )
            response["saf"] = {
                "base_co2_kg": savings.base_co2_kg,
                "saf_co2_kg": savings.saf_co2_kg,
                "co2_reduction_kg": savings.co2_reduction_kg,
                "reduction_percentage": savings.reduction_percentage,
            }
        return response
    except ValueError as e:
        raise HTTPException(status_code=400, detail=str(e))
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))


@app.get("/synthetic-route")
async def synthetic_route(
    from_code: str = Query(..., alias="from"),
    to_code: str = Query(..., alias="to"),
    waypoints: int = Query(12, ge=0, le=100),
    speed_kts: float = Query(450.0, ge=100, le=1000),
    _auth: None = Depends(_verify_api_key),
):
    try:
        route = generate_route(
            from_code,
            to_code,
            num_waypoints=waypoints,
            speed_kts=speed_kts,
        )
        return {
            "from": route.origin.as_dict(),
            "to": route.destination.as_dict(),
            "total_distance_km": route.total_distance_km,
            "total_time_hours": route.total_time_hours,
            "waypoints": [
                {
                    "name": wp.name,
                    "latitude": wp.latitude,
                    "longitude": wp.longitude,
                    "distance_from_origin_km": wp.distance_from_origin_km,
                }
                for wp in route.waypoints
            ],
        }
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))


def run_server(host: str = "0.0.0.0", port: int = 8000):
    import uvicorn

    uvicorn.run(app, host=host, port=port)


if __name__ == "__main__":
    run_server()



from __future__ import annotations

from collections.abc import Iterable, Sequence
from dataclasses import dataclass
from typing import Optional

from ..core.airports import get as get_airport
from ..core.loader import get_all_airports
from ..models.airport import Airport
from ..utils.spatial_index import build_spatial_index


@dataclass(frozen=True)
class HubScore:
    airport: Airport
    hub_score: float
    connectivity_score: float
    type_score: float


def _airport_type_score(airport: Airport) -> float:
    if airport.type == "large_airport":
        return 1.0
    if airport.type == "medium_airport":
        return 0.6
    if airport.type == "small_airport":
        return 0.3
    return 0.1


class NetworkIntelligence:
    def __init__(
        self,
        airports: Optional[Sequence[Airport]] = None,
        radius_km: float = 500.0,
        max_neighbors: int = 100,
    ) -> None:
        if airports is None:
            airports = get_all_airports()
        self.airports = list(airports)
        self.radius_km = radius_km
        self.max_neighbors = max_neighbors
        self._index = build_spatial_index(self.airports)

    def _connectivity_score(self, airport: Airport) -> float:
        neighbors = self._index.within_radius(
            airport.latitude_deg,
            airport.longitude_deg,
            self.radius_km,
        )
        count = max(len(neighbors) - 1, 0)
        normalized = min(count / float(self.max_neighbors), 1.0)
        return normalized

    def score_airport(self, airport: Airport) -> HubScore:
        type_score = _airport_type_score(airport)
        connectivity_score = self._connectivity_score(airport)
        hub_score = (0.6 * connectivity_score) + (0.4 * type_score)
        if airport.scheduled_service:
            hub_score = min(hub_score + 0.05, 1.0)

        return HubScore(
            airport=airport,
            hub_score=hub_score,
            connectivity_score=connectivity_score,
            type_score=type_score,
        )

    def rank_airports(self, airports: Optional[Iterable[Airport]] = None) -> list[HubScore]:
        if airports is None:
            airports = self.airports
        scored = [self.score_airport(a) for a in airports]
        return sorted(scored, key=lambda s: s.hub_score, reverse=True)


def identify_global_hubs(
    top_n: int = 10,
    airports: Optional[Sequence[Airport]] = None,
    scheduled_service_only: bool = True,
) -> list[HubScore]:
    if airports is None:
        airports = get_all_airports()

    filtered = [
        a
        for a in airports
        if a.type in ("large_airport", "medium_airport")
        and (a.scheduled_service or not scheduled_service_only)
    ]

    intelligence = NetworkIntelligence(filtered)
    return intelligence.rank_airports(filtered)[:top_n]


def hub_intelligence_score(code: str) -> HubScore:
    airport = get_airport(code, code_type="auto")
    if airport is None:
        raise ValueError(f"Airport not found: {code}")

    intelligence = NetworkIntelligence()
    return intelligence.score_airport(airport)



"""Tests for AeroNavX REST API — auth, rate limiting, and core endpoints."""

import os
from unittest.mock import patch

import pytest
from fastapi.testclient import TestClient


@pytest.fixture(autouse=True)
def _reset_api_state():
    """Reset module-level state between tests."""
    import importlib

    import aeronavx.api.server as srv_module

    # Clear rate limit store
    srv_module._rate_limit_store.clear()
    yield


def _make_client(api_key=None, rate_limit=None):
    """Create a fresh TestClient with optional env overrides."""
    env = {}
    if api_key is not None:
        env["AERONAVX_API_KEY"] = api_key
    if rate_limit is not None:
        env["AERONAVX_RATE_LIMIT"] = str(rate_limit)

    with patch.dict(os.environ, env, clear=False):
        import importlib

        import aeronavx.api.server as srv_module

        # Re-read env vars
        srv_module._API_KEY = os.environ.get("AERONAVX_API_KEY")
        srv_module._RATE_LIMIT = int(os.environ.get("AERONAVX_RATE_LIMIT", "60"))
        srv_module._rate_limit_store.clear()

        return TestClient(srv_module.app)


class TestHealthEndpoint:
    def test_health_no_auth(self):
        """Health endpoint is always open, even with API key set."""
        client = _make_client(api_key="secret123")
        resp = client.get("/health")
        assert resp.status_code == 200
        data = resp.json()
        assert data["status"] == "ok"
        assert "version" in data

    def test_health_no_key_configured(self):
        client = _make_client()
        resp = client.get("/health")
        assert resp.status_code == 200


class TestApiKeyAuth:
    def test_open_access_without_key(self):
        """When no API key is configured, endpoints are open."""
        client = _make_client()
        resp = client.get("/airport/IST")
        assert resp.status_code == 200

    def test_valid_key_allows_access(self):
        """Correct API key grants access."""
        client = _make_client(api_key="mysecret")
        resp = client.get("/airport/IST", headers={"X-API-Key": "mysecret"})
        assert resp.status_code == 200

    def test_invalid_key_returns_401(self):
        """Wrong API key returns 401."""
        client = _make_client(api_key="mysecret")
        resp = client.get("/airport/IST", headers={"X-API-Key": "wrong"})
        assert resp.status_code == 401

    def test_missing_key_returns_401(self):
        """Missing API key returns 401 when key is configured."""
        client = _make_client(api_key="mysecret")
        resp = client.get("/airport/IST")
        assert resp.status_code == 401


class TestRateLimiting:
    def test_rate_limit_enforced(self):
        """Exceeding rate limit returns 429."""
        client = _make_client(rate_limit=5)
        for i in range(5):
            resp = client.get("/airport/IST")
            assert resp.status_code == 200, f"Request {i + 1} should succeed"
        resp = client.get("/airport/IST")
        assert resp.status_code == 429

    def test_health_exempt_from_rate_limit(self):
        """Health endpoint is exempt from rate limiting."""
        client = _make_client(rate_limit=2)
        for _ in range(5):
            resp = client.get("/health")
            assert resp.status_code == 200


class TestCoreEndpoints:
    def test_airport_found(self):
        client = _make_client()
        resp = client.get("/airport/JFK")
        assert resp.status_code == 200
        assert "iata_code" in resp.json()

    def test_airport_not_found(self):
        client = _make_client()
        resp = client.get("/airport/ZZZZ99")
        assert resp.status_code in (400, 404)

    def test_distance(self):
        client = _make_client()
        resp = client.get("/distance", params={"from": "IST", "to": "JFK"})
        assert resp.status_code == 200
        assert "distance" in resp.json()

    def test_nearest(self):
        client = _make_client()
        resp = client.get("/nearest", params={"lat": 41.0, "lon": 29.0, "n": 3})
        assert resp.status_code == 200
        assert "airports" in resp.json()

    def test_search(self):
        client = _make_client()
        resp = client.get("/search", params={"q": "Istanbul"})
        assert resp.status_code == 200



from __future__ import annotations

import os
from pathlib import Path
from typing import Any, Literal, Optional

from .config import CACHE_DIR, EMBED_MODEL, HF_TOKEN, USE_LOCAL_ONLY
from .index import AirportIndex
from .utils import try_import_pandas

ReturnFormat = Literal["auto", "dataframe", "list"]


def _load_sentence_transformer(
    model_name: str,
    cache_dir: Optional[Path],
    token: Optional[str],
    local_only: bool,
):
    try:
        from sentence_transformers import SentenceTransformer
    except ImportError as exc:
        raise ImportError(
            "Semantic search requires sentence-transformers. Install with "
            "`pip install aeronavx[hf]`."
        ) from exc

    if local_only:
        os.environ.setdefault("HF_HUB_OFFLINE", "1")

    if token and not os.getenv("HUGGINGFACE_HUB_TOKEN"):
        os.environ["HUGGINGFACE_HUB_TOKEN"] = token

    kwargs: dict[str, Any] = {}
    if cache_dir is not None:
        kwargs["cache_folder"] = str(cache_dir / "hf")
    if local_only:
        kwargs["local_files_only"] = True
    if token:
        kwargs["token"] = token

    try:
        return SentenceTransformer(model_name, **kwargs)
    except TypeError:
        kwargs.pop("token", None)
        kwargs.pop("local_files_only", None)
        return SentenceTransformer(model_name, **kwargs)
    except Exception as exc:
        message = str(exc).lower()
        if local_only:
            message = (
                f"Offline mode enabled and model '{model_name}' "
                "is not available in the local cache. "
                "Disable offline mode or pre-download the model."
            )
            raise RuntimeError(message) from exc
        if token and ("401" in message or "unauthorized" in message or "forbidden" in message):
            raise RuntimeError(
                "Hugging Face token is invalid or lacks access to the requested model."
            ) from exc
        if "404" in message or ("not found" in message and "model" in message):
            raise RuntimeError(f"Hugging Face model '{model_name}' was not found.") from exc
        if "connection" in message or "timed out" in message or "network" in message:
            raise RuntimeError(
                "Unable to download model. Check network access or enable offline mode "
                "with a cached model."
            ) from exc
        raise


class SemanticAirportSearch:
    def __init__(
        self,
        airports: Any,
        model_name: str = EMBED_MODEL,
        *,
        cache_dir: Optional[Path] = CACHE_DIR,
        token: Optional[str] = HF_TOKEN,
        local_only: bool = USE_LOCAL_ONLY,
        model: Optional[Any] = None,
        use_faiss: bool = True,
    ) -> None:
        if cache_dir is not None and not isinstance(cache_dir, Path):
            cache_dir = Path(cache_dir).expanduser()

        if model is None:
            model = _load_sentence_transformer(model_name, cache_dir, token, local_only)

        self.model = model
        self.model_name = model_name
        self.cache_dir = cache_dir
        self.index = AirportIndex(
            airports,
            model,
            model_name,
            cache_dir=cache_dir,
            use_faiss=use_faiss,
        )

    def search(
        self,
        query: str,
        top_k: int = 5,
        return_format: ReturnFormat = "auto",
    ):
        results = self.index.search(query, top_k=top_k)

        if return_format == "list":
            return results

        pandas = try_import_pandas()

        if return_format == "dataframe" and pandas is None:
            raise ImportError("pandas is required for return_format='dataframe'.")

        if return_format == "auto" and pandas is None:
            return results

        df = pandas.DataFrame(results)
        return df



import os

import pytest

from aeronavx.hf.semantic_search import SemanticAirportSearch
from aeronavx.models.airport import Airport

pytest.importorskip("sentence_transformers")
pytest.importorskip("numpy")


RUN_REAL = os.getenv("AERONAVX_RUN_REAL_MODEL_TESTS") == "1"


@pytest.mark.skipif(not RUN_REAL, reason="Set AERONAVX_RUN_REAL_MODEL_TESTS=1 to run.")
def test_semantic_search_real_model_rankings(tmp_path):
    airports = [
        Airport(
            id=1,
            ident="LTFM",
            type="large_airport",
            name="Istanbul Airport",
            latitude_deg=41.275278,
            longitude_deg=28.751944,
            elevation_ft=163,
            continent="AS",
            iso_country="TR",
            iso_region="TR-34",
            municipality="Istanbul",
            scheduled_service=True,
            gps_code="LTFM",
            iata_code="IST",
            local_code=None,
            home_link=None,
            wikipedia_link=None,
            keywords="Istanbul international main airport",
        ),
        Airport(
            id=2,
            ident="LTSA",
            type="large_airport",
            name="Istanbul Sabiha Gokcen International Airport",
            latitude_deg=40.8986,
            longitude_deg=29.3092,
            elevation_ft=312,
            continent="AS",
            iso_country="TR",
            iso_region="TR-34",
            municipality="Istanbul",
            scheduled_service=True,
            gps_code="LTSA",
            iata_code="SAW",
            local_code=None,
            home_link=None,
            wikipedia_link=None,
            keywords="Sabiha Gokcen Istanbul airport",
        ),
        Airport(
            id=3,
            ident="KJFK",
            type="large_airport",
            name="John F Kennedy International Airport",
            latitude_deg=40.639801,
            longitude_deg=-73.7789,
            elevation_ft=13,
            continent="NA",
            iso_country="US",
            iso_region="US-NY",
            municipality="New York",
            scheduled_service=True,
            gps_code="KJFK",
            iata_code="JFK",
            local_code=None,
            home_link=None,
            wikipedia_link=None,
            keywords="New York international airport",
        ),
        Airport(
            id=4,
            ident="KLGA",
            type="large_airport",
            name="LaGuardia Airport",
            latitude_deg=40.7769,
            longitude_deg=-73.874,
            elevation_ft=21,
            continent="NA",
            iso_country="US",
            iso_region="US-NY",
            municipality="New York",
            scheduled_service=True,
            gps_code="KLGA",
            iata_code="LGA",
            local_code=None,
            home_link=None,
            wikipedia_link=None,
            keywords="New York domestic airport",
        ),
        Airport(
            id=5,
            ident="KEWR",
            type="large_airport",
            name="Newark Liberty International Airport",
            latitude_deg=40.6895,
            longitude_deg=-74.1745,
            elevation_ft=18,
            continent="NA",
            iso_country="US",
            iso_region="US-NJ",
            municipality="Newark",
            scheduled_service=True,
            gps_code="KEWR",
            iata_code="EWR",
            local_code=None,
            home_link=None,
            wikipedia_link=None,
            keywords="New York area Newark airport",
        ),
        Airport(
            id=6,
            ident="RJAA",
            type="large_airport",
            name="Narita International Airport",
            latitude_deg=35.7647,
            longitude_deg=140.3864,
            elevation_ft=141,
            continent="AS",
            iso_country="JP",
            iso_region="JP-12",
            municipality="Tokyo",
            scheduled_service=True,
            gps_code="RJAA",
            iata_code="NRT",
            local_code=None,
            home_link=None,
            wikipedia_link=None,
            keywords="Tokyo Narita international airport",
        ),
        Airport(
            id=7,
            ident="RJTT",
            type="large_airport",
            name="Tokyo Haneda Airport",
            latitude_deg=35.552258,
            longitude_deg=139.779694,
            elevation_ft=21,
            continent="AS",
            iso_country="JP",
            iso_region="JP-13",
            municipality="Tokyo",
            scheduled_service=True,
            gps_code="RJTT",
            iata_code="HND",
            local_code=None,
            home_link=None,
            wikipedia_link=None,
            keywords="Tokyo Haneda airport",
        ),
        Airport(
            id=8,
            ident="EGLL",
            type="large_airport",
            name="London Heathrow Airport",
            latitude_deg=51.4775,
            longitude_deg=-0.461389,
            elevation_ft=83,
            continent="EU",
            iso_country="GB",
            iso_region="GB-ENG",
            municipality="London",
            scheduled_service=True,
            gps_code="EGLL",
            iata_code="LHR",
            local_code=None,
            home_link=None,
            wikipedia_link=None,
            keywords="London Heathrow airport",
        ),
        Airport(
            id=9,
            ident="EGKK",
            type="large_airport",
            name="London Gatwick Airport",
            latitude_deg=51.1481,
            longitude_deg=-0.1903,
            elevation_ft=203,
            continent="EU",
            iso_country="GB",
            iso_region="GB-ENG",
            municipality="London",
            scheduled_service=True,
            gps_code="EGKK",
            iata_code="LGW",
            local_code=None,
            home_link=None,
            wikipedia_link=None,
            keywords="London Gatwick airport",
        ),
    ]

    searcher = SemanticAirportSearch(
        airports,
        cache_dir=tmp_path,
        token=os.getenv("HF_TOKEN")
        or os.getenv("HF_API_TOKEN")
        or os.getenv("HUGGINGFACE_HUB_TOKEN"),
    )

    queries = {
        "Istanbul international airport": "IST",
        "New York airport": "JFK",
        "Tokyo Narita": "NRT",
        "London Heathrow": "LHR",
    }

    for query, expected_iata in queries.items():
        results = searcher.search(query, top_k=3, return_format="list")
        iatas = [r["iata"] for r in results]
        scores = [r["score"] for r in results]

        assert iatas[0] == expected_iata
        assert expected_iata in iatas
        assert scores == sorted(scores, reverse=True)
        assert scores[0] - scores[-1] >= 0.01

        repeated = searcher.search(query, top_k=3, return_format="list")
        assert [r["iata"] for r in repeated] == iatas



from aeronavx.core import loader


def test_load_airports_reapplies_filters(tmp_path):
    csv_content = "\n".join(
        [
            "id,ident,type,name,latitude_deg,longitude_deg,elevation_ft,continent,iso_country,iso_region,municipality,scheduled_service,gps_code,iata_code,local_code,home_link,wikipedia_link,keywords",
            "1,AAA1,large_airport,Alpha Airport,0.0,0.0,0,NA,XX,XX-1,Alpha City,yes,AAA1,AAA,,,,",
            "2,BBB1,small_airport,Beta Airport,1.0,1.0,0,NA,XX,XX-1,Beta City,no,BBB1,BBB,,,,",
        ]
    )
    csv_path = tmp_path / "airports.csv"
    csv_path.write_text(csv_content)

    loader.load_airports(data_path=csv_path, force_reload=True)

    filtered = loader.load_airports(data_path=csv_path, include_types=["large_airport"])

    assert len(filtered) == 1
    assert filtered[0].iata_code == "AAA"

    loader.clear_cache()



import pytest

from aeronavx.core.synthetic_routes import generate_route
from aeronavx.models.airport import Airport


def _airport(code, lat, lon):
    return Airport(
        id=None,
        ident=code,
        type="large_airport",
        name=f"{code} Airport",
        latitude_deg=lat,
        longitude_deg=lon,
        elevation_ft=None,
        continent=None,
        iso_country=None,
        iso_region=None,
        municipality=None,
        scheduled_service=True,
        gps_code=code,
        iata_code=code,
        local_code=None,
        home_link=None,
        wikipedia_link=None,
        keywords=None,
    )


def test_generate_route_basic():
    origin = _airport("AAA", 0.0, 0.0)
    destination = _airport("BBB", 10.0, 10.0)

    route = generate_route(origin, destination, num_waypoints=5)

    assert route.total_distance_km > 0
    assert route.total_time_hours > 0
    assert len(route.waypoints) == 5


def test_generate_route_zero_waypoints():
    origin = _airport("AAA", 0.0, 0.0)
    destination = _airport("BBB", 5.0, 5.0)

    route = generate_route(origin, destination, num_waypoints=0)
    assert len(route.waypoints) == 0


def test_generate_route_invalid_waypoints():
    origin = _airport("AAA", 0.0, 0.0)
    destination = _airport("BBB", 5.0, 5.0)

    with pytest.raises(ValueError):
        generate_route(origin, destination, num_waypoints=-1)



from __future__ import annotations

from pathlib import Path
from typing import Any, Optional

from .utils import sanitize_model_id

CACHE_VERSION = "v1"


def embeddings_cache_path(cache_dir: Path, model_id: str, signature: str) -> Path:
    safe_model = sanitize_model_id(model_id)
    return cache_dir / "hf" / "semantic_search" / safe_model / f"{signature}.npz"


def load_embeddings(path: Path) -> Optional[tuple[Any, str, int, str]]:
    if not path.exists():
        return None

    import numpy as np

    try:
        data = np.load(path, allow_pickle=False)
        embeddings = data["embeddings"]
        signature = str(data["signature"].item())
        count = int(data["count"].item())
        version = str(data["version"].item())
    except Exception:
        return None

    if getattr(embeddings, "shape", None) is None:
        return None
    if embeddings.shape[0] != count:
        return None

    return embeddings, signature, count, version


def save_embeddings(path: Path, embeddings: Any, signature: str) -> None:
    path.parent.mkdir(parents=True, exist_ok=True)

    import numpy as np

    np.savez_compressed(
        path,
        embeddings=embeddings,
        signature=signature,
        count=embeddings.shape[0],
        version=CACHE_VERSION,
    )



"""Runway loading and management for AeroNavX."""

import csv
from pathlib import Path
from typing import Optional, Union

from ..exceptions import DataLoadError
from ..models.runway import Runway
from ..utils.logging import get_logger

logger = get_logger()

_runways: list[Runway] = []
_airport_index: dict[str, list[Runway]] = {}  # airport_ident -> runways
_loaded = False


def _find_runway_file() -> Path:
    """Find the runways CSV file."""
    possible_paths = [
        Path(__file__).parent.parent / "data" / "runways.csv",
        Path(__file__).parent.parent.parent / "data" / "ourairports_runways.csv",
        Path.cwd() / "data" / "runways.csv",
        Path.cwd() / "data" / "ourairports_runways.csv",
    ]

    for path in possible_paths:
        if path.exists():
            return path

    raise DataLoadError(
        f"Could not find runways.csv. Searched in: {', '.join(str(p) for p in possible_paths)}"
    )


def _parse_bool(value: str) -> Optional[bool]:
    """Parse boolean value from CSV."""
    if not value or value.lower() in ("", "no", "0", "false"):
        return False
    if value.lower() in ("yes", "1", "true"):
        return True
    return None


def _parse_int(value: str) -> Optional[int]:
    """Parse integer value from CSV."""
    try:
        return int(value) if value else None
    except ValueError:
        return None


def _parse_float(value: str) -> Optional[float]:
    """Parse float value from CSV."""
    try:
        return float(value) if value else None
    except ValueError:
        return None


def load_runways(
    data_path: Optional[Union[Path, str]] = None,
    force_reload: bool = False,
) -> list[Runway]:
    """
    Load runways from CSV file.

    Args:
        data_path: Path to runways CSV file (default: auto-detect)
        force_reload: Force reload even if already loaded

    Returns:
        List of Runway objects

    Example:
        >>> runways = load_runways()
        >>> print(f"Loaded {len(runways):,} runways")
        Loaded 47,373 runways
    """
    global _runways, _airport_index, _loaded

    if _loaded and not force_reload:
        return _runways

    if data_path is None:
        data_path = _find_runway_file()
    else:
        data_path = Path(data_path)
        if not data_path.exists():
            raise DataLoadError(f"Runway file not found: {data_path}")

    logger.info(f"Loading runways from {data_path}")

    runways = []
    skipped = 0

    try:
        with open(data_path, encoding="utf-8") as f:
            reader = csv.DictReader(f)

            for row in reader:
                try:
                    airport_ident = row.get("airport_ident", "").strip()
                    if not airport_ident:
                        skipped += 1
                        continue

                    runway = Runway(
                        id=_parse_int(row.get("id", "")) or 0,
                        airport_ref=_parse_int(row.get("airport_ref", "")) or 0,
                        airport_ident=airport_ident,
                        length_ft=_parse_float(row.get("length_ft", "")),
                        width_ft=_parse_float(row.get("width_ft", "")),
                        surface=row.get("surface", "").strip() or None,
                        lighted=_parse_bool(row.get("lighted", "")),
                        closed=_parse_bool(row.get("closed", "")),
                        le_ident=row.get("le_ident", "").strip() or None,
                        le_latitude_deg=_parse_float(row.get("le_latitude_deg", "")),
                        le_longitude_deg=_parse_float(row.get("le_longitude_deg", "")),
                        le_elevation_ft=_parse_float(row.get("le_elevation_ft", "")),
                        le_heading_degT=_parse_float(row.get("le_heading_degT", "")),
                        le_displaced_threshold_ft=_parse_float(
                            row.get("le_displaced_threshold_ft", "")
                        ),
                        he_ident=row.get("he_ident", "").strip() or None,
                        he_latitude_deg=_parse_float(row.get("he_latitude_deg", "")),
                        he_longitude_deg=_parse_float(row.get("he_longitude_deg", "")),
                        he_elevation_ft=_parse_float(row.get("he_elevation_ft", "")),
                        he_heading_degT=_parse_float(row.get("he_heading_degT", "")),
                        he_displaced_threshold_ft=_parse_float(
                            row.get("he_displaced_threshold_ft", "")
                        ),
                    )

                    runways.append(runway)

                except Exception as e:
                    logger.debug(f"Skipping runway due to error: {e}")
                    skipped += 1
                    continue

    except Exception as e:
        raise DataLoadError(f"Failed to load runways: {e}")

    _runways = runways
    _build_indices()
    _loaded = True

    logger.info(f"Loaded {len(_runways):,} runways (skipped {skipped})")

    return _runways


def _build_indices() -> None:
    """Build indices for fast runway lookups."""
    global _airport_index

    _airport_index.clear()

    for runway in _runways:
        if runway.airport_ident not in _airport_index:
            _airport_index[runway.airport_ident] = []
        _airport_index[runway.airport_ident].append(runway)


def get_runways_by_airport(airport_ident: str) -> list[Runway]:
    """
    Get all runways for a specific airport.

    Args:
        airport_ident: Airport identifier (IATA, ICAO, or ident code)

    Returns:
        List of Runway objects for the airport

    Example:
        >>> runways = get_runways_by_airport("KJFK")
        >>> for rwy in runways:
        ...     print(f"{rwy.designation}: {rwy.length_ft:.0f}ft")
        04L/22R: 14511ft
        04R/22L: 11351ft
        13L/31R: 10000ft
        13R/31L: 12079ft
    """
    if not _loaded:
        load_runways()

    # Try exact match first
    if airport_ident in _airport_index:
        return _airport_index[airport_ident]

    # Try uppercase
    airport_ident_upper = airport_ident.upper()
    if airport_ident_upper in _airport_index:
        return _airport_index[airport_ident_upper]

    return []


def get_longest_runway(airport_ident: str) -> Optional[Runway]:
    """
    Get the longest runway for an airport.

    Args:
        airport_ident: Airport identifier

    Returns:
        Longest Runway object, or None if no runways found

    Example:
        >>> runway = get_longest_runway("KJFK")
        >>> print(f"{runway.designation}: {runway.length_ft:.0f}ft")
        04L/22R: 14511ft
    """
    runways = get_runways_by_airport(airport_ident)
    if not runways:
        return None

    return max(
        (r for r in runways if r.length_ft is not None),
        key=lambda r: r.length_ft or 0,
        default=None,
    )


def get_paved_runways(airport_ident: str) -> list[Runway]:
    """
    Get all paved runways for an airport.

    Args:
        airport_ident: Airport identifier

    Returns:
        List of paved Runway objects

    Example:
        >>> runways = get_paved_runways("KJFK")
        >>> print(f"{len(runways)} paved runways")
        4 paved runways
    """
    runways = get_runways_by_airport(airport_ident)
    return [r for r in runways if r.is_paved]


def clear_cache() -> None:
    """Clear runway data cache."""
    global _runways, _airport_index, _loaded

    _runways.clear()
    _airport_index.clear()
    _loaded = False

    logger.info("Cleared runway data cache")



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Optional, Union

from ..core.airports import get as get_airport
from ..core.timezone import get_timezone_offset
from ..exceptions import AirportNotFoundError
from ..models.airport import Airport


class TravelDirection(Enum):
    EASTWARD = "eastward"
    WESTWARD = "westward"
    NONE = "none"


class JetLagSeverity(Enum):
    MILD = "mild"
    MODERATE = "moderate"
    SEVERE = "severe"


@dataclass(frozen=True)
class JetLagResult:
    timezone_difference_hours: float
    direction: TravelDirection
    severity: JetLagSeverity
    estimated_recovery_days: float


def _direction(diff_hours: float) -> TravelDirection:
    if diff_hours > 0:
        return TravelDirection.EASTWARD
    if diff_hours < 0:
        return TravelDirection.WESTWARD
    return TravelDirection.NONE


def _severity(diff_hours: float) -> JetLagSeverity:
    abs_diff = abs(diff_hours)
    if abs_diff <= 2:
        return JetLagSeverity.MILD
    if abs_diff <= 5:
        return JetLagSeverity.MODERATE
    return JetLagSeverity.SEVERE


def _recovery_days(diff_hours: float, age: int) -> float:
    base_days = abs(diff_hours) / 2.0
    if age >= 50:
        base_days *= 1.2
    elif age <= 18:
        base_days *= 0.8
    return max(base_days, 0.0)


def _resolve_airport(value: Union[Airport, str], label: str) -> Airport:
    if isinstance(value, Airport):
        return value
    if isinstance(value, str):
        airport = get_airport(value, code_type="auto")
        if airport is None:
            raise AirportNotFoundError(f"{label} airport not found: {value}")
        return airport
    raise TypeError(f"{label} must be an Airport instance or airport code string.")


def calculate_jet_lag(
    origin: Union[Airport, str],
    destination: Union[Airport, str],
    age: int = 30,
    origin_offset: Optional[float] = None,
    destination_offset: Optional[float] = None,
) -> JetLagResult:
    if age <= 0:
        raise ValueError("age must be positive")

    origin_airport = _resolve_airport(origin, "origin")
    destination_airport = _resolve_airport(destination, "destination")

    if origin_offset is None:
        origin_offset = get_timezone_offset(origin_airport)
    if destination_offset is None:
        destination_offset = get_timezone_offset(destination_airport)

    if origin_offset is None or destination_offset is None:
        raise ValueError("Could not determine timezone offsets for jet lag calculation.")

    diff_hours = destination_offset - origin_offset
    direction = _direction(diff_hours)
    severity = _severity(diff_hours)
    recovery = _recovery_days(diff_hours, age)

    return JetLagResult(
        timezone_difference_hours=diff_hours,
        direction=direction,
        severity=severity,
        estimated_recovery_days=recovery,
    )



from .core.airports import get as get_airport
from .core.airports import get_by_iata, get_by_icao
from .core.airports import search_by_name as search_airports_by_name
from .core.distance import distance, distance_km, distance_mi, distance_nmi
from .core.emissions import estimate_co2_kg_by_codes as estimate_co2_kg_for_segment
from .core.emissions_advanced import (
    AircraftType,
    EmissionsResult,
    FuelType,
    SafSavings,
    calculate_flight_emissions,
    compare_saf_savings,
)
from .core.geodesy import great_circle_path, initial_bearing, midpoint
from .core.network_intelligence import (
    HubScore,
    NetworkIntelligence,
    hub_intelligence_score,
    identify_global_hubs,
)
from .core.passenger_experience import (
    JetLagResult,
    JetLagSeverity,
    TravelDirection,
    calculate_jet_lag,
)
from .core.routing import estimate_flight_time_hours as estimate_flight_time
from .core.routing import route_distance
from .core.runways import (
    get_longest_runway,
    get_paved_runways,
    get_runways_by_airport,
)
from .core.search import airports_within_radius, nearest_airport, nearest_airports
from .core.statistics import (
    get_continent_stats,
    get_country_stats,
    get_global_stats,
    get_top_countries_by_airports,
    get_top_countries_by_large_airports,
)
from .core.synthetic_routes import (
    SyntheticRoute,
    Waypoint,
    generate_route,
    generate_route_by_codes,
)
from .core.timezone import get_timezone_offset
from .core.weather import get_metar, get_taf
from .exceptions import (
    AeroNavXError,
    AirportNotFoundError,
    DataLoadError,
    InvalidAirportCodeError,
    RoutingError,
    WeatherDataError,
)
from .models import Airport, Runway

_semantic_engine = None


def semantic_search(query: str, top_k: int = 5, return_format: str = "auto"):
    global _semantic_engine

    if _semantic_engine is None:
        try:
            from .hf.semantic_search import SemanticAirportSearch
        except ImportError as exc:
            raise ImportError(
                "Semantic search requires the HF extra. Install with " "`pip install aeronavx[hf]`."
            ) from exc

        from .core.loader import get_all_airports

        airports = get_all_airports()
        _semantic_engine = SemanticAirportSearch(airports)

    return _semantic_engine.search(query, top_k=top_k, return_format=return_format)


def clear_semantic_search_cache() -> None:
    global _semantic_engine
    _semantic_engine = None


__version__ = "3.1.0"

__all__ = [
    "Airport",
    "Runway",
    "get_airport",
    "get_by_iata",
    "get_by_icao",
    "distance",
    "distance_km",
    "distance_mi",
    "distance_nmi",
    "initial_bearing",
    "midpoint",
    "great_circle_path",
    "nearest_airport",
    "nearest_airports",
    "airports_within_radius",
    "estimate_flight_time",
    "route_distance",
    "estimate_co2_kg_for_segment",
    "calculate_flight_emissions",
    "compare_saf_savings",
    "AircraftType",
    "FuelType",
    "EmissionsResult",
    "SafSavings",
    "calculate_jet_lag",
    "JetLagResult",
    "JetLagSeverity",
    "TravelDirection",
    "NetworkIntelligence",
    "identify_global_hubs",
    "hub_intelligence_score",
    "HubScore",
    "generate_route",
    "generate_route_by_codes",
    "SyntheticRoute",
    "Waypoint",
    "search_airports_by_name",
    "get_metar",
    "get_taf",
    "get_timezone_offset",
    "get_runways_by_airport",
    "get_longest_runway",
    "get_paved_runways",
    "get_country_stats",
    "get_continent_stats",
    "get_global_stats",
    "get_top_countries_by_airports",
    "get_top_countries_by_large_airports",
    "semantic_search",
    "clear_semantic_search_cache",
    "AeroNavXError",
    "AirportNotFoundError",
    "InvalidAirportCodeError",
    "DataLoadError",
    "RoutingError",
    "WeatherDataError",
]
